:- module('Core', [exports/1]).

:- use_module(library(lists),         [reverse/2] ).
:- use_module(library(dynamic),       [assert/1, clause/2]).
:- use_module(library(aggregates),    [findall/3]).
:- use_module(library(iso_byte_char), [char_code/2] ).
:- use_module(library(terms),         [atom_concat/2] ).

%:- use_package(trace).

% :- include(ast_ops).
% Should update each time we update ast_ops.

% Abstree operators.
:- op(500, xfx, :=).
:- op(550, xfx, @).

% This operator is used by core and trans_lazy.

:- op(500, xfx, #).

% Type operators.

:- op(950, xfy, ->).
:- op(960, xfy, =>).

% Only necessary if we have free vars in types.
type_arity(X, 0) :-
	var(X),
	!.

type_arity(_A -> B, N) :-
	!,
	type_arity(B, N1),
	N is N1 + 1.
	
type_arity(_, 0) :-
	!.

:- initialization(create_core_exports).

:- dynamic       exports/1, tonf/2, tohnf/2, is_constructor/1.
:- multifile     tohnf/2, tonf/2, is_constructor/1.
:- discontiguous tohnf/2, tonf/2, is_constructor/1.

%% Exports of core

create_core_exports :-
	% Core types.
	%retractall(exports(_)),
	export_type(qid('Prelude', list(A)), 
	             [(qid('Prelude', []),  qid('Prelude', list(_))),
		      (qid('Prelude', ':'), A -> qid('Prelude', list(A)) -> qid('Prelude', list(A)))]),
        export_type(qid('Prelude', tuple(A, B)), 
	             [(qid('Prelude', ','), A -> B -> qid('Prelude', tuple(A, B))) ] ),
	export_type(qid('Prelude', 'Bool'),      
	             [(qid('Prelude', 'True'),  qid('Prelude', 'Bool')), 
		      (qid('Prelude', 'False'), qid('Prelude', 'Bool'))]), 
	export_type(qid('Prelude', 'Int'),  []),
	export_type(qid('Prelude', 'Char'), []),
	!.
	%write(exito),
	%nl.

	
%% Translation of constructors.

%% Basically this is a copy of trans_lazy. For each cons we generate:
%%  - constructor(Cons).
%%  - tohnf(Constructor(Params), Constructor(Params)).
%%  - tohnf(Constructor(P1..Pn), Constructor(T1..Tn)) :-
%%           tonf(P1, R1),
%%           ..
%%           tonf(Pn, Rn).


print_generated :-
	findall(A, is_constructor(A), L),
	write(L),
	nl,
	findall((A, B,C), clause(tonf(A,B),C), L2),
	write(L2).

export_type(Name, ConsL) :-
	assert(exports(type(Name, ConsL))),
	!,
	translate_cons(ConsL).

translate_cons([]).
translate_cons([C|CL]) :-
	translate_constructor(C),
	translate_cons(CL).

new_vars(0, []) :- !.
new_vars(N, [_|Ts]) :-
	N1 is N-1,
	new_vars(N1, Ts).

make_name(qid('Prelude', '[]'), []).
make_name(qid('Prelude', ':'),  .).	
%make_name(qid('Prelude', ','),  ',').

make_name(qid(Module, Name), SName) :-
	atom_concat([Module, '_', Name], SName).

translate_constructor((QName, Type)) :-
	!,
	make_name(QName, Name),
	assert(is_constructor(Name)),
	type_arity(Type, Arity),
	cons_tohnf(Name, Arity),
	cons_tonf(Name, Arity).

cons_tohnf(Name, Arity) :-
	new_vars(Arity, Vars),
	NCons =.. [Name|Vars],
	assert((tohnf(NCons, NCons) :- !)).

cons_tonf(Name, Arity) :-
	new_vars(Arity, Vars1),
	copy_term(Vars1, Vars2),
	cons_tonf_list(Vars1, Vars2, Body),
	NCons1 =.. [Name|Vars1],
	NCons2 =.. [Name|Vars2],
	assert((tonf(NCons1, NCons2) :- Body)).

cons_tonf_list([], [], (!)).
cons_tonf_list([L1|LL1], [L2|LL2], (Tonf1, TonfL)) :-
	Tonf1 = tonf(L1, L2),
	!,
	cons_tonf_list(LL1, LL2, TonfL).
	
% Head normal form detection.


is_in_hnf(A) :-
	var(A),
	!.

% Should use constant detection.
is_in_hnf(const(_)) :-
	!.

is_in_hnf(A) :-
	number(A),
	!.

%
is_in_hnf([])  :- !.
is_in_hnf([_|_]) :-
	!.

% Constructor form.
is_in_hnf(Term) :-
	functor(Term, C, _),
	is_constructor(C),
	!.

%% Ugly workaround.
%% Should Fix with a new constructor framework.
is_in_hnf(Term#Term) :-
	functor(Term, C, _),
	is_constructor(C),
	!.

%% Head Normal Form translation.

%% Be carefull as we can have some infite terms here:
%% as tonhf(main#R, R).

%% Basic tohnf
%% Already in hnf.
tohnf(A, B) :- 
	is_in_hnf(A),
	!,
	A = B.

% Lazy eval.
tohnf(_#A, B) :-
        \+ var(A),
	!,
	A = B.

%% --- Converts a character into its ASCII value.
%% ord :: Char -> Int
%% ord external
tohnf('Prelude_ord'(Char)#R, R) :-
	!,
	char_code(Char, R).

%% --- Converts an ASCII value into a character.
%% chr :: Int -> Char
%% chr external
tohnf(chr(Int)#R, R) :-
	!, 
	char_code(R, Int).

%% Integer operations.
tohnf('Prelude_+'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	B is F+G.

tohnf('Prelude_-'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	B is F-G.

tohnf('Prelude_*'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	B is F*G.

tohnf('Prelude_div'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	B is F//G.

tohnf((D mod E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	B is (F mod G).

%% Comparison operators.

% Builting arithmetic, belongs in the prelude?
tohnf((D<E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	(
	    F < G -> 
	    B = 'True' 
	; 
	    B = 'False'
	).

tohnf((D>E)#B,B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	(
	    F > G -> 
	    B = 'True' 
	; 
	    B = 'False'
	).

tohnf((D>=E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	(
	    F >= G -> 
	    B = 'True' 
	; 
	    B = 'False'
	).

tohnf((D=<E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	(
	    F =< G -> 
	    B = 'True' 
	; 
	    B = 'False'
	).
 
%% Lazy application.

tohnf(apply_list(A,B)#C, C) :-
        apply_list_lazy(A, B, D),
        eval_lazy(D, C).

eval_lazy(pap(A,0,B)#_, C) :- 
	!,
        reverse(B, D),
        E=..[A|D], % Tambien puede haber constr. currif. (sin puntero)
        (
	    is_constructor(A)->
	    F =.. [tohnf,E,C]
	;
	    F=..[tohnf,E#C,C]
	),
        call(F).

eval_lazy(A, A).

apply_lazy(pap(A,B,C)#X, D, pap(A,E,[D|C])#X) :-
        E is B-1.

apply_list_lazy(A, [], A).
apply_list_lazy(A, [B|C], D) :-
        apply_lazy(A, B, E),
        apply_list_lazy(E, C, D).

%% Constraint support

%% First support for LHSE:

tohnf(conditions([])#'Prelude_False', 'False') :-
	throw(condition_error).

tohnf(conditions([C|CL])#R, R) :-
	(
	    tohnf(C, R) ->
	    true
	;
	    tohnf(conditions(CL)#R, R)
	).

tohnf(cond(C,E)#R,R) :- 
	!,
	tonf(C, Rc),
	Rc == 'Prelude_True',
	tohnf(E,R).

%% Old constraint support must verify
tohnf(constraint([])#'Success', 'Success') :-
	!.
tohnf(constraint([C|Cs])#'Success','Success') :-
	!,
	tohnf(C,'Success'),
	tohnf(constraint(Cs)#_,'Success').

tohnf(choice([(constraint(Const)#_->Expr)#_|Cs])#R,R) :-
	!,
	(
	    tohnf(constraint(Const)#A,A) ->
	    tohnf(Expr,R) 
	; 
	    tohnf(choice(Cs)#R,R)
	).


% Not operator.
tohnf(not(A)#B, B) :-
        tonf(A, X),
	(
 	    X = false ->
	    B = true
	;
	    B = false
 	).


% tonf rules.
tonf(A, B) :-
	var(A),
	!,
	A = B.

tonf(A, B) :-
	number(A),
	!, 
	A = B.

%tonf(A, B) :-
%	A = const(_),
%       !, A = B.					
         
tonf(_A#B, B) :-
	\+ var(B),
	!.

tonf(eq(A,B),'True') :- 
	!,
	tohnf(A,At),
	tohnf(B,Bt),
	eq(At,Bt).

tonf([], []).
tonf([A|B], [AN|BN]) :-
	tonf(A, AN),
	tonf(B, BN).

tonf(A#B, C) :-
        tohnf(A#B, D),
	!,
        tonf(D, C).



%%%% Igualdad

tohnf((A =:= B)#'Success', 'Success') :-
	tohnf(A,A1),
	tohnf(B,B1),
	eq(A1,B1).

tohnf('Prelude_=='(A,B)#C,C) :- % Igualdad estricta (==)
	tohnf(A,A1),
	tohnf(B,B1),
	!,
	'=='(A1,B1,C).

'=='(A, B, 'Prelude_True') :-
	eq(A, B),
	!.

'=='(A, B, 'Prelude_False') :-
	\+ eq(A, B),
	!.


% eq(A,B):-(var(A);var(B)),!,A=B.
eq(A,B) :-
	var(A),
	!,
	tonf(B,A).

eq(A,B) :-
	var(B),
	!,
	tonf(A,B).

eq([],[]) :- 
	!.
eq([H|Ts], [A|Zs]) :- 
	!,
	tohnf(H,H1),
	tohnf(A,A1),
	eq(H1,A1),
	tohnf(Ts,Ts1),
	tohnf(Zs,Zs1),
	eq(Ts1,Zs1).

eq(A,B) :-
  A =.. [Ca | Alist],
  B =.. [Cb | Blist],
  !,
  Ca = Cb,
  eq(Alist, Blist).

eq(A, A) :-
	!.


eq_list([],[]).
eq_list([A|As], [B|Bs]) :-
	eq(A, B),
	eq_list(As, Bs).

and_bool(A, C, B) :-
        tonf(A, X),
        (
	    X == 'False' ->
	    B = 'False'
	;
	    tonf(C,B)
	).

or_bool(A, C, B) :-
        tonf(A, X),
        (
	    X == 'True' ->
	    B =  'True'
	;
	    tonf(C, B)
	).

%% Deprecated and to be revised code:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% tohnf_list([],[]).
% tohnf_list([A|As],[B|Bs]) :-
% 	tohnf(A,B),
% 	tohnf_list(As,Bs).


% Monads belong in the prelude.
%tohnf(done#done,done) :- !.
% tohnf(return(A)#B,B) :- !, tohnf(A,B).

% tohnf(getChar#char(X), char(X)) :-
% 	!,
% 	get0(X).

% tohnf(putChar(char(C))#X, Y) :- 
% 	!,
% 	X = done,
% 	Y = done,
% 	put(C).

% tohnf((A>>B)#C,C) :- 
% 	!,
% 	tohnf(A,_),
% 	tohnf(B,C).

% tohnf(>>=(A,B)#C,C) :- 
% 	!,
% 	tohnf(A,Ahnf),
% 	B = pap(Right,Remain,List)#_,
% 	tohnf(apply_list(pap(Right,Remain,List)#_,[Ahnf])#C,C).

% This belongs in the prelude too.
% tohnf(and_bool(A,C)#B, B) :-
% 	and_bool(A,C,B).

% tohnf(or_bool(A,C)#B, B) :-
% 	or_bool(A,C,B).


%%%%%%%%%%%%%%%%%%%%%% Ejecucion del obj en batch: %%%%%%%%%%%%%%%%%

%filter_variables_batch([],[]) :-
% 	!.
% filter_variables_batch([loc(_)-_|Vs],Vfs) :-
% 	!,
% 	filter_variables_batch(Vs,Vfs).

% filter_variables_batch([Var-_|Vs],Vfs) :- % Quitar las var. Rxxx
% 	name(Var,[82|_]), 
% 	!,
% 	filter_variables_batch(Vs,Vfs).

% filter_variables_batch([Var-Val|Vs],[Var is Val|Vfs]) :-
% 	filter_variables_batch(Vs,Vfs).

% unfree_vars([],_).

% unfree_vars([Atom|Vs],Atom) :-
% 	unfree_vars(Vs,Atom).

% mark_free_vars([]) :- 
% 	!.
% mark_free_vars([Item|Items]) :-
% 	Item =.. [_,L,R],
% 	(
% 	    var(R) ->
% 	    R = L 
% 	; 
% 	    true
% 	),
% 	mark_free_vars(Items).

% mark_block_vars([]) :- 
% 	!.
% mark_block_vars([L-_|Items]) :-
% 	unfree_vars(L,'<blocking>'),
% 	mark_block_vars(Items).

% exec_batch(Goal, Dicc) :-
% 	prompt(_,''),
% 	Goal = tonf(_Exp,R1),
% 	varset(Dicc,VarsGoal),
% 	call_residue(Goal,Res), %comentar?
% 						%  call(Goal),%descomentar?
%         mark_block_vars(Res),                   %comentar?
% 	mark_free_vars(Dicc),                
% 	varset(R1,VarsResult),
% 	unfree_vars(VarsResult,'<free>'),
% 	write('Result: '),write(R1),nl,
% 	(
% 	    VarsGoal == [] ->
% 	    true
% 	;
% 	    write('Substitution: '),
% 	    write(Dicc),
% 	    nl,
% 	    nl,
% 	    write('More?'),
% 	    get(Answer),
% 	    nl,
% 	    \+ (Answer = 0'; )
%         ).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

