:- module(debug, [dump_abstree/1, dump_deftrees/1, pretty_list_write/1, diff_lists/3, 
	          debug_write/1, debug_write_string/1] ).

:- use_module(library(lists),      [difference/3, append/3]).
:- use_module(library(strings),    [write_string/2] ).
:- use_module(library(hiordlib),   [map/3] ).
:- use_module(library(aggregates), [findall/3]).

% Library modules.
:- use_module(symtab,  [clean_symtab/0, get_rules/2, get_object_property/3] ).
%:- use_module(deftree, [call_defTree/2] ).
:- use_module(deftree).
% For more pretty printing, we include operators.
:- include(ast_ops).
:- include(type_ops).

% Debug write
debug_write(E) :-
	write(user_error, E).

debug_write(pos(L, C)) :-
	write(user_error, 'Line: '),
	write(user_error, L),
	write(user_error, ' Col: '),
	write(user_error, C).

debug_write_string(S) :-
	write_string(user_error, S).

dump_rules(Rules) :-
	findall(rule(N, R), get_rules(N, R), Rules),
	!.

dump_symtab(Entries) :-
	findall(entry(N, P, V), get_object_property(N, P, V), Entries),
	!.	

dump_deftrees(DefTrees) :-
	findall(deftree(N, Dt), (
				    get_rules(N, R),
				    \+ R = [builtin(_)],
				    call_defTree(N, Dt)
				), 
				DefTrees),
	!.
				    

pretty_write(rule(N,R)) :-
	pretty_write(N),
	display(': '),
	display('\n'),
	pretty_list_write(R),
	display('\n').

pretty_write(entry(N, P, V)) :-
	pretty_write(N),
	display('.'),
	write(P),
	display(': '),
	write(V),
	display('\n').

pretty_write(qid(Module, Name)) :-
	write(Module),
	write('.'),
	write(Name).

pretty_write(plus(E)) :-
 	write('+ '),
	pretty_write(E).
%%	write('\n').

pretty_write(minus(E)) :-
 	write('- '),
	pretty_write(E).
%%	write('\n').

pretty_write(deftree(N, D)) :-
	debug_write('\n\nDeftree for '),
	debug_write(N),
	debug_write('\n-----------------\n\n'),
	debug_write_deftree(D, 0).

pretty_write(A := B) :-
	pretty_write(A),
	write(' := ' ),
	pretty_write(B),
	write('\n').

pretty_write(E@Params) :-
	pretty_write(E),
	write('@['),
	pretty_list_write_deco(Params),
	write(']').

pretty_write(E) :-
	!,
	write(E).

%% This is very wrong, we should fix this.
indent(0).
indent(N) :- 
	debug_write(' '),
	N1 is N - 1,
	indent(N1).

debug_write_deftree(fallo-[], N) :-
	indent(N),
	debug_write('Fallo en deftree'),
	debug_write('\n'),
	!.

debug_write_deftree(regla-[Rule], N) :-
	indent(N),
	debug_write(Rule),
	debug_write('\n'),
	!.

debug_write_deftree((Cons, Dt), N) :-
	indent(N),
	debug_write(Cons),
	debug_write(' => '),
	debug_write_deftree(Dt, N),
	!.

debug_write_deftree(branch(Pivot, SubTrees)-_Ns, N) :-
	debug_write('Branch by '),
	debug_write(Pivot),
	debug_write('\n'),
	N1 is N + 4,
	debug_write_deftrees(SubTrees, N1),
	%debug_write_deftree(NS),
	debug_write('\n'),
	!.

debug_write_deftree(or(SubTrees)-_Ns, N) :-
	debug_write('Or branch'),
	debug_write('\n'),
	N1 is N + 4,
	debug_write_deftrees(SubTrees, N1),
	%debug_write_deftree(NS),
	debug_write('\n'),
	!.

debug_write_deftrees([], _).
debug_write_deftrees([D|DL], N) :-
	debug_write_deftree(D, N),
	debug_write('\n'),
	debug_write_deftrees(DL, N).

pretty_list_write([]).
pretty_list_write([E|ES]) :-
	pretty_write(E),
	pretty_list_write(ES).


pretty_list_write_deco([]).
pretty_list_write_deco([E|ES]) :-
	pretty_write(E),
	(
	    \+ ES = [],
	    write(', '),
	    pretty_list_write_deco(ES)
	;
	    true
	).


% Debug function, writes symbol table to file.
dump_abstree(Abstree) :-
	dump_rules(Rules),
	dump_symtab(St),
	append(Rules, St, Abstree),
	!.


%%% This predicate generates a unified diff style list difference.

mark(E, M, R) :-
	R =.. [M|[E]].

diff_lists(L1, L2, LR) :-
	difference(L1, L2, Minus),
	difference(L2, L1, Plus),
	map(Minus, mark(minus), Min),
	map(Plus,  mark(plus),  Plu),
	append(Min, Plu, LR).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

