:- module(parser, [parse/1],[]).

%%

:- use_module(error,        [do_error/1] ).
:- use_module(lexer_compat, [init_scanner/1, token_have/1, token_look/1,
	                     token_skip/0, last_token/1, token_back/0]).
:- use_module(symtab,       [add_rule/2, get_object_property/3, set_object_property/3, add_constructor/2] ).
:- use_module(debug,        [debug_write/1, debug_write_string/1] ).
:- use_module(types,        [type_arity/2] ).
:- use_module(cmdopts,      [cmd_options/2] ).
:- use_module(modules,      [start_module/1, end_module/1, add_exports/1, load_module/1, current_mod_name/1] ).
:- use_module(priotree,     [pt_insert/3] ).
:- use_module(scope,        [open_scope/1, close_scope/0, insert_symbol/2, lookup_symbol_unref/2] ).

%% Ciao libraries.
:- use_module(library(dynamic),    [assert/1] ).
:- use_module(library(lists),      [append/3, length/2] ).
:- use_module(library(write),      [write/1] ).

%% We use some high-order parsing.
:- use_package(hiord).

%% Operators includes.
:- include(ast_ops).
:- include(type_ops).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Module support.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<module>' :-
	(
	    token_have(lex_reserved(module)),
	    token_have(lex_conid(ModName)),
	    start_module(ModName),
	    '<module_exports>',
	    token_have(lex_reserved(where)),
	    open_scope(ModName),
	    insert_symbol(':', qid('Prelude', ':')),
	    (
		\+ cmd_options('-np', _) ->
		load_module(module('Core',    [], none)),
		load_module(module('Prelude', [], none))
	    ;
		true
	    ),
	    token_have('{'),
	    '<block>',
	    token_have('}'),
	    %close_scope,
	    end_module(ModName),
	    token_have(eof)
	;
	    throw_parser_error('Error in module declaration')
	).	

'<module_exports>' :-
	(
	    token_look('(') ->
	    open_close_scan(Exports, list_scan('<module_export_unit>', ','), '(', ')'),
	    add_exports(Exports)	
        ;
	    add_exports(all)
        ).

%% A module export unit can be: (Prelude pg 25)
%% - function ids.
%% - datatypes with or without constructors.
%% - another modules.

'<module_export_unit>'(id(X)) :-
	token_have(lex_id(X)).

'<module_export_unit>'(type(Name, ListOfCons)) :-
	token_have(lex_conid(Name)),
	open_close_scan(ListOfCons, '<conid>', '(', ')').

'<conid>'(X) :-
	token_have(lex_conid(X)).

%% Import declarations.
'<import_decl>'(module(Name, Symbols, Modifier)) :-
	(
	    \+ '<module_import_mod>'(Modifier) ->
	    Modifier = none
	),
	'<conid>'(Name),
	(
	    token_look(';') ->
	    Symbols = []
	;
	    open_close_scan(Symbols, list_scan('<module_export_unit>', ','), '(', ')')
	).

'<module_import_mod>'(qualified) :-
	token_have(lex_reserved(qualified)).

'<module_import_mod>'(hiding) :-	
	token_have(lex_reserved(hiding)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% A block is a list of block_unit.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<block>' :-
	decl_list('<block_unit>').

%%  block_unit ::=  
%%       import module_import
%%       data type_declaration
%%       fixity fixity_arguments
%%       function_name '::' type_expr
%%       function_name equat

'<block_unit>' :-
	token_have(lex_reserved(import)),
	!,
	'<import_decl>'(Imports),
	load_module(Imports).

'<block_unit>' :-
	token_have(lex_reserved(data)),
	!,
	'<type_declaration>'(_Name, _Type).

% We need typer support for this.
% '<block_unit>' :-
% 	token_have(lex_reserved(type)),
% 	!,
% 	'<type_alias>'(Name, Type),
% 	make_type_alias(Name, Type).

'<block_unit>' :-
	token_look(lex_reserved(FixityKeyword)),
	member(FixityKeyword, [infix,infixl,infixr]),
	!,
	token_skip,
	'<fixity_argument>'(Nat, Ops),
	record_fixity(FixityKeyword, Nat, Ops).

'<block_unit>':-
	(
	    '<function_name>'(FunctionName)
	;
	    NT = yes
	),
	(
	    token_have(lex_op('::')) ->
	    '<type_expr>'(TypeExpr),
	    record_type(FunctionName, TypeExpr)
	;
	    %% Care here with scopes.
	    token_have(lex_reserved(external)),
	    FunctionName = qid(_, Name),
	    insert_symbol(Name, FunctionName),
	    add_rule(FunctionName, builtin(external))
	;
	    (
		var(NT) ->
		token_back
	    ;
		true
	    ),
	    '<equat>'(Equat),
	    %% FIXME: This is an ugly hack.
	    Equat = (Fname@_) := _,
	    record_equat(Fname, Equat)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Symbol names, must add qualified names.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This is ugly, should cleanup.
'<function_name>'(OpName):-
	token_have('('),
	(
	    token_look(lex_op(OpName1)) ->
	    token_skip,
	    token_have(')'),
	    mod_name(OpName1, OpName)
	;
	    token_back,
	    !,
	    fail
	).

'<function_name>'(FunctionName):-
	token_have(lex_id(FunctionName1)),
	mod_name(FunctionName1, FunctionName).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Reglas de declaracion de tipos:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type_expr ::=
%      type_expr_unit 
%      type_expr_unit '->') type_expr
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<type_expr>'(TypeExpr):-
	'<type_expr_unit>'(Tu),
	(
	    token_have(lex_op(->)) ->
	    '<type_expr>'(Te),
	    TypeExpr = (Tu -> Te)
	;
	    TypeExpr = Tu
	).

% type_expr_unit representa la declaracin de un tipo.
% type_expr_unit ::=
%      Constructor {TypeExprUnit}          : Tipo existente mas nmero de argumentos adecuado a su aridad.
%      id                                  : typevars
%      ()                                  : Unit type.
%      '(' (type_expr) [','type_expr]  ')' : Tuplas
%      '[' type_expr  ']'                  : Listas
%      Bool | String | Char | Int | Float  : Tipos bsicos, deberan ir en el preludio.

'<type_expr_unit>'(TypeExpr) :- % Regla 1.
	'<conid>'(C),
	!,
	'<type_exprs_unit>'(TypeExprList),
	( 
	    TypeExprList = [] ->
	    TypeExpr = C
	;
	    TypeExpr =.. [C,TypeExprList]
	).

%% egallego@babel:
%% Type variables must be seen as vars in the typer.

'<type_expr_unit>'(typevar(I)) :- % Regla 2
	token_have(lex_id(I)),
	!.

'<type_expr_unit>'(TypeExpr):-
	token_have('('),
	!,
	(
	    token_have(')') ->
	    TypeExpr = unit
	;
	    (
		'<type_expr_list>'(TypeListAux),
		(
		    length(TypeListAux, 1),
		    TypeListAux = [A] ->
		    TypeExpr=A
		;
		    TypeExpr = tuple(TypeListAux)
		),
		token_mustbe(')',') expected')
	    )
	).

'<type_expr_unit>'([TypeExpr]) :-
	open_close_scan(TypeExpr, '<type_expr>', '[', ']').

'<type_exprs_unit>'([T1|Ts]) :-       % <type_exprs_unit>::=
	'<type_expr_unit>'(T1),
	!,                            % lambda | <type_expr_unit><type_exprs_unit>
	'<type_exprs_unit>'(Ts).

'<type_exprs_unit>'([]).
'<type_expr_list>'([T1|Ts]) :- % <type_expr_list>::=
	'<type_expr>'(T1),    % <type_expr> | <type_expr>,<type_expr_list>
	(
	    token_have(',') ->
	    '<type_expr_list>'(Ts)
	;
	    Ts = []
	).

'<type_exprs>'([T1|Ts]) :-       % <type_exprs>::=
	'<type_expr>'(T1),
	!,                       % lambda | <type_expr><type_exprs>
	'<type_exprs>'(Ts).

'<type_exprs>'([]).


% type_declaration ::=
%       CONID type_var_ids '=' constr_declaration ['|' constr_declaration]
%

'<type_declaration>'(Name, Type):-
	token_mustbe(lex_conid(Cname), 'A constructor name was expected'),
	'<type_var_ids>'(Ids),
	UName =.. [Cname|Ids],
	mod_name(UName, Name),
	token_mustbe(lex_op('='),      '= expected after type variables'),
	'<constr_declaration>'(ConstrDeclaration),
	(
	    token_have(lex_op('|')) -> 
	    '<constr_declarations>'(ConstrDecs),
	    Type = [ConstrDeclaration|ConstrDecs]
	;
	    Type = [ConstrDeclaration]
	),
	addconstructors(Type, Name).

% Type vars:
%
% type_var_ids ::=
%      { ID }

'<type_var_ids>'(List):-!,
	(
	    token_have(lex_id(Id)) ->
	    (
		'<type_var_ids>'(Ids),
		List = [typevar(Id)|Ids]
	    )
	;
	    List = []
	).  

'<constr_declarations>'(ConstDeclarations):-
	'<constr_declaration>'(ConstDec),
	(
	    token_have(lex_op('|')) ->
	    (
		'<constr_declarations>'(ConstDecs),
		ConstDeclarations = [ConstDec|ConstDecs]
	    )
	;
	    (
		ConstDeclarations = [ConstDec]
	    )
	).

% Declaracin de los constructores del tipo de datos.
%
% constr_declaration ::= 
%       CONID type_exprs
%      

'<constr_declaration>'(ConstrDeclaration):-
	token_mustbe(lex_conid(Cname),'A constructor name was expected'),
	mod_name(Cname, QCname),
	'<type_exprs>'(TypeExprs),
	ConstrDeclaration = (QCname, TypeExprs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Expressions: %%%%%%%%%%%%%%%%%%%%%%%%%%%%
% expr ::=
%     '\' pattern '->' expr
%     IF expr THEN expr ELSE expr
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


'<expr>'(lambda(P,Expr)) :-
	token_have('\\'),
	!,
	'<pattern>'(P),
	token_mustbe(lex_op(->),'-> expected in lambda expresion'),
	'<expr>'(Expr).

'<expr>'(qid('Prelude', if_then_else)@[If,Then,Else]) :-
	token_have(lex_reserved(if)),
	!,
	'<expr>'(If),
	token_mustbe(lex_reserved(then),'then keyword expected'),
	'<expr>'(Then),
	token_mustbe(lex_reserved(else),'else keyword expected'),
	'<expr>'(Else).

% '<expr'(let(BindingsL, Expr)) :-
% 	token_have(lex_reserved(let)),
% 	!,
% 	list_scan(BindingsL, '<
	

'<expr>'(Expr):-
	'<func_expr>'(E1),
	'<infix_exp>'(E1, Expr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Normal infix expressions.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<infix_exp>'(OExp, Exp) :-
	(
	    ( %% Bad bad operators.
		token_look(lex_op(Op)),
		\+ forbidden_in_exp(Op),
		token_skip  
	    ) ->
	    '<func_expr>'(FExp),
	    lookup_symbol_unref(Op, Qop),
	    pt_insert(OExp, (Qop, FExp), NExp),
	    '<infix_exp>'(NExp, Exp)	    
	;
	    OExp = Exp
	).

forbidden_in_exp(Op) :-
	member(Op, ['=','|']).

%% Applicative expressions.
'<func_expr>'(Expr):-
	seq_scan([Function|Params], '<basic_expr>'),
	(
	    %% Constant basic_expressions doesn't want to be applied an empty list.
	    Params = [],
	    (
		(
		    Function = id(_)
		;
		    Function = lambda(_, _)
		) ->
		Expr = Function@Params
	    ;
		Expr = Function
	    )
        ; 
	    Expr = Function@Params
        ).

%% Basic expression support.

'<basic_expr>'(Id) :-
	token_have(lex_id(X)),
	!,
	lookup_symbol_unref(X, Id).

'<basic_expr>'(Cons) :-
	token_have(lex_conid(C)), 
	!,
	lookup_symbol_unref(C, Cons).

'<basic_expr>'(const(num, N)) :-
	token_have(lex_num(N)),
	!.

'<basic_expr>'(Expr):-
	token_have('('),
	!,
	(
	    token_have(')') ->
	    Expr = qid('Prelude', '()' )
	;
	    token_have(lex_op(X)),
	    token_have(')'),
	    lookup_symbol_unref(X, Expr)
	;
	    list_scan(InnerExpr, '<expr>', ','),
	    (
		list_to_tuple(InnerExpr, Expr)
	    ),
	    token_have(')')
	).

%% FIXME, must support list comprehension.
'<basic_expr>'(Expr) :-
	token_have('['),
	!,
	(
	    token_have(']') ->
	    Expr = qid('Prelude', '[]')
	;
	    list_scan(ExprS, '<expr>', ','),
	    desugar_list(ExprS, Expr),
	    token_have(']')
	).

%% Helper predicates.

list_to_tuple([A] , A) :- !.
list_to_tuple([L|LL], qid('Prelude', ',')@[L,B]) :-
	list_to_tuple(LL, B).

desugar_list([], qid('Prelude', '[]')).
desugar_list([A|B], qid('Prelude', ':')@[A, BS]) :-
	desugar_list(B, BS).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Functions declaration.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<equat>'(Equat):-
	'<funlhs>'(Pattern),
	(
	    token_have(lex_op(=)) -> 
	    '<expr>'(Expr),
	    Equat = Pattern := Expr
	;
	    token_look(lex_op('|')) -> 
	    '<cond_expr>'(CondExpr),
	    Equat = Pattern := CondExpr
	;		
	    derror(pipe_expected)
	),
	close_scope.

'<funlhs>'(Pattern) :-
	'<simple_pattern>'(P1),
	%% We look for a operator.
	(
	    token_look(lex_op(Op)),
	    \+ Op = '=' ->
	    token_skip,
	    '<simple_pattern>'(P2),
	    mod_name(Op, Qop),
	    Pattern = Qop@[P1, P2],
	    insert_symbol(Op, Qop),
	    open_scope(Qop)
	;
	    P1 = param(F),
	    mod_name(F, QF) ->
	    (
		seq_scan(PatList, '<simple_pattern>') ->
		Pattern = QF@PatList
	    ;
		Pattern = QF@[]
	    ),
	    insert_symbol(F, QF),
	    open_scope(QF)
	;
	    derror(function_decl_error)
	),
	insert_scope_args(Pattern).

insert_scope_args(param(X)) :- 
	insert_symbol(X, param(X)).

insert_scope_args(_Cons@Patt) :-
	insert_scope_args_list(Patt).

insert_scope_args([L]) :-
	insert_scope_args(L).

insert_scope_args((A,B)) :-
	insert_scope_args(A),
	insert_scope_args(B).

insert_scope_args(_).

insert_scope_args_list([]).
insert_scope_args_list([E|EL]) :-
	insert_scope_args(E),
	insert_scope_args_list(EL).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Patterns.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<simple_pattern>'(param(X)) :-
	token_have(lex_id(X)).	

'<simple_pattern>'(Cons) :-
	token_have(lex_conid(X)),
	!,
	lookup_symbol_unref(X, Cons).

%% Fix for constant framework.
'<simple_pattern>'(const(num, X)) :-
	token_have(lex_num(X)).

'<simple_pattern>'(Tuple) :-
	token_have('('),
	(
	    token_have(')') ->
	    Tuple = qid('Prelude', '()')
	;
	    list_scan(PatList, '<pattern>', ','),
	    token_have(')'),
	    (
		PatList = [L] ->
		Tuple = L
	    ;
		list_to_tuple(PatList, Tuple)
	    )
	).

'<simple_pattern>'(qid('Prelude', '[]')) :-
	token_have('['),
	(
	    token_have(']') ->
	    true
	;
	    token_back,
	    fail
	).
		
'<simple_pattern>'(PatList) :-
	open_close_scan(PatListS, list_scan('<pattern>', ','), '[', ']'),
	desugar_list(PatListS, PatList).


'<pattern>'(Pattern) :-
	'<head_pattern>'(Head),
	(
	    token_have(lex_op(':')) ->
	    '<pattern>'(Tail),
	    Pattern = qid('Prelude', ':')@[Head, Tail]
	;
	    Pattern = Head
	).

'<head_pattern>'(QConid@Body) :-
	token_have(lex_conid(Conid)),
	lookup_symbol_unref(Conid, QConid),
	seq_scan(Body, '<pattern>').

'<head_pattern>'(Pattern) :-
	'<simple_pattern>'(Pattern).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Right hand side constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
'<cond_expr>'(conditions([C|CL])) :-
	token_have(lex_op('|')),
	'<expr>'(CExpr),
	token_mustbe(lex_op(=),'= expected in constraint'),
	'<expr>'(EBody),
	C = (CExpr -> EBody),
	(
	    token_look(lex_op('|')) ->
	    '<cond_expr>'(conditions(CL))
	;
	    CL = []
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Operators declaration.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<fixity_argument>'(Nat, Ops):-
	token_have(lex_num(Nat)),
	'<infix_op_ids>'(Ops).

'<infix_op_ids>'([op(Op)|Ops]):-
	token_mustbe(lex_op(Op),'Malformed infix operator'),
	(
	    token_have(',') -> 
	    '<infix_op_ids>'(Ops) 
	; 
	    Ops = []
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Grammar end.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% High order predicates to parse common patterns.
%% FIXME, we should add error handling here.

:- meta_predicate open_close_scan(?, pred(1), ?, ?).

%% Has use with *Tok free too.
open_close_scan(Res, Goal, OTok, CTok) :-
	token_have(OTok),
	!,
	Goal(Res),
	token_have(CTok).

:- meta_predicate list_scan(?, pred(1), ?).

list_scan([Res|ResL], Goal, Symbol) :-
	Goal(Res),
	(
	    token_have(Symbol) ->
	    list_scan(ResL, Goal, Symbol)
	;
	    ResL = []
	).	

% Sequences (with no commas between elements).
:- meta_predicate seq_scan(?, pred(1)).

seq_scan([Res|ResL], Goal) :-
	Goal(Res),
	(
	    seq_scan(ResL, Goal),
	    !
	;
	    ResL = []
	).

%% We don't use list_scan here as we want different error handling.
:- meta_predicate decl_list(pred(0)).
decl_list(Goal) :-
	token_have(';'),
	!,
	decl_list(Goal).

decl_list(Goal) :-
	Goal,
	!,
	(
	    token_have(';') ->
	    decl_list(Goal)
	;
	    true
	).

%% Utilities for recording things in the symbol table.

% Add a list of cons to symbol table.

addconstructors([], _).

addconstructors([(Cons, TypeOfCons) | R], TC) :-
	(
	    TypeOfCons = [] ->
	    NewTypeC = TC
	;
    	    append(TypeOfCons, [TC], TypeOfCons1),
	    list2type(TypeOfCons1, NewTypeC)
	),
	Cons = qid(_, Name),
	insert_symbol(Name, Cons),
%	write(symbol_insertion(Name, Cons)),
	nl,
	add_constructor(TC, Cons),
	set_object_property(Cons, type, NewTypeC),
	set_object_property(Cons, constructs, TC),
	addconstructors(R, TC).


list2type([A], A).
list2type([A|LR], A -> TR) :-
	list2type(LR, TR).

record_type(Name,Type) :-
	(
	    get_object_property(Name, type, _) 	->
	    derror(dup_type(Name))
	;
	    set_object_property(Name, type, Type)
	).

record_equat(Name, Eq) :-
	check_arity(Name, Eq), 
	add_rule(Name, Eq).
	

%% Esto corresponde al tipador.
%% De momento exigimos que tengan la declaracin de tipos hecha, en otro caso 
%% error.

%% Por cierto el nmero de argumentos tiene que ser fijo, puede que haya que 
%% pensar esto ms.

check_arity(Name, _Def) :-
	(	
	    get_object_property(Name, type, _Type)
	;
	    throw(type_error('We have no typer at the moment, sorry', Name))
	).

%% Recording of new operators.
record_fixity(_Fix, _Nat, []).
record_fixity(Fixity, Nat, [op(Op)|Rs]) :-
	mod_name(Op, QOp),
	set_object_property(QOp, priority, Nat),
	set_object_property(QOp, assoc,    Fixity),
	record_fixity(Fixity, Nat, Rs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Main parsing function.

parse(File):-
	
	catch(init_scanner(File), lexer_error(E), do_lexer_error(E)), 
	catch('<module>', parser_error(T, E, R), do_parser_error(T, E, R)).


%% For future shell
% parse_expr(Expr) :-
% 	catch(init_scanner_string(Expr), E, do_error(E)),
% 	'<expr>'(Expr).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Error handling.

%% Dummy error function.
derror(Text) :-
	throw_parser_error(Text).

%% Parser recovery. 

%% Resync looks for the token especified and then continues parsing.
do_recovery(resync(Token, Parse)) :-
	resynch(Token),
	call(Parse).

do_recovery(no_recovery) :-
	throw(error('Unrecoverable parser error')).

resynch(Token) :-
	(
	    token_look(Token) ->
	    true
	;
	    (
		token_look(eof) ->
		last_token(T),
		throw(parser_error(T, 'End of file', no_recovery))
	    ;
		token_skip,
		resynch(Token)
	    )
	).

%% New error format
%% This needs a big cleanup.

do_parser_error(Token, Error, Recovery) :-
	Token = (T, Pos),
	debug_write('Parser error in: '),
	debug_write(Pos),
	debug_write('\nPossible token: '),
	debug_write(T),
	debug_write('\nError: '),
	debug_write(Error),
	debug_write('\n'),
	do_recovery(Recovery).

throw_parser_error(Error) :-
	last_token([Token|_]),
	throw(parser_error(Token, Error, resync(';', '<block>'))).

do_lexer_error((Pos, CandToken, Input, _NewInput)) :-
	debug_write('Lexer error in: '),
	debug_write(Pos),
	debug_write('\nPossible token: '),
	debug_write(CandToken),
	debug_write('\nInput: '),
	debug_write_string(Input),
	debug_write('\n').	


token_mustbe(Token, ErrCode) :-
	(
	    token_have(Token),
	    !
	; 
	    derror(ErrCode)
	). 

mod_name(Name1, qid(M, Name1)) :-
	current_mod_name(M).

% start_function(Name, Params) :-
% 	mod_name(Name, Mname),
% 	insert_symbol(Name, Mname),
% 	insert_params(Params),
% 	start_scope(Name).

% insert_params([]).
% insert_params([P|PL]) :-
% 	insert_symbol(P, param(P)),
% 	insert_params(PL).
	

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

