:- module(symtab, [clean_symtab/0, add_rule/2, get_rules/2, update_rules/2, 
	           set_object_property/3, get_object_property/3, 
		   add_constructor/2, get_constructors/2,
		   delete_property/2, replace_object_property/3,
		   load_interfaces/1, object_arity/2]).

% This module implements a function oriented symbol table.

:- use_module(library(dynamic),    [assert/1, retract/1, retractall/1]).

% 

:- use_module(types,  [ type_arity/2 ]).
:- use_module(scope,  [ insert_symbol/2 ]).

% Private dynamics

:- dynamic func_rules/2.
:- dynamic object_properties/3.
:- dynamic type_cons/2.

% Current scope(module, symbol).
:- dynamic current_scope/2.

clean_symtab :-
	retractall(func_rules(_, _) ),
	retractall(object_properties(_, _, _) ),
	retractall(type_cons(_, _) ).

% New symtab begins here

% Rule handling predicates.
add_rule(Name, AbsTree) :-
	retract(func_rules(Name, L)),
	assert(func_rules(Name, [AbsTree|L])).

% If we don't have any rules, add the first.
add_rule(Name, AbsTree) :-
	assert(func_rules(Name, [AbsTree])).

update_rules(Name, Rules) :-
	retract(func_rules(Name, _)),
	assert(func_rules(Name, Rules)).

get_rules(Name, ListOfAbsTrees) :-
	func_rules(Name, ListOfAbsTrees).

% General property handling.
set_object_property(Name, Property, Value) :-
	check_property(Property),
	assert(object_properties(Name, Property, Value)).

get_object_property(Name, Property, Value) :-
%	check_property(Property),
	object_properties(Name, Property, Value).

delete_property(Name, Property) :-
	retractall(object_properties(Name, Property, _)).

replace_object_property(Name, Property, Value) :-
	delete_property(Name, Property),
	set_object_property(Name, Property, Value).

% Function properties are well defined and symbol
% table enforces this rectriction.
is_property(type).

is_property(exports).
is_property(imports).

%% Operators priority.
is_property(priority).
is_property(assoc).

% FIXME: provisional until I see if can replace link in the whole Sloth.
is_property(constructs).

% If P is not a property throw and error.
check_property(P) :-
	is_property(P),
	!.

check_property(P) :-
	throw(invalid_property(P)).

%% object_arity returns the expected number or args that a cons or pred
%% should take.
object_arity(Object, Arity) :-
	get_object_property(Object, type, Type),
	type_arity(Type, Arity),
	!.

% Type database:
% 
% A data type in Curry usually has some constructors.
%
% Data type -> List of Constructors.
% Data type -> arity, etc...
%
% Constructor -> Arity, Type
%
add_constructor(Type, Cons) :-
	retract(type_cons(Type, ConsL)),
	assert(type_cons(Type, [Cons|ConsL])).

add_constructor(Type, Cons) :-
	assert(type_cons(Type, [Cons])).

get_constructors(Type, ListOfCons) :-
	type_cons(Type, ListOfCons).

%% Fill symbol table with a module's interface.
load_interfaces([]).
load_interfaces([I|IF]) :-
	load_interface(I),
	!,
	load_interfaces(IF).

load_interface(iface(id(Symbol, Type, OpAttr))) :-
	add_rule(Symbol, builtin(imported)),
	set_object_property(Symbol, type, Type),
	( 
	    OpAttr = op(Assoc, Prio) ->
	    set_object_property(Symbol, assoc, Assoc),
	    set_object_property(Symbol, priority, Prio)
	;
	    true
	),
	qid(_, BareSymbol) = Symbol,
	insert_symbol(BareSymbol, Symbol).

load_interface(iface(type(Name, ConsL))) :-
	load_cons(Name, ConsL).

load_cons(_, []).
load_cons(Type, [(CName, CType)|R]) :-
	add_constructor(Type, CName),
	set_object_property(CName, type, CType),
	set_object_property(CName, constructs, Type),
	load_cons(Type, R),
	qid(_, BareSymbol) = CName,
	insert_symbol(BareSymbol, CName).


%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

